.. _photom_selectors:

Reference Selection Keywords for PHOTOM
+++++++++++++++++++++++++++++++++++++++
CRDS selects appropriate PHOTOM reference based on the following keyword.
PHOTOM is not applicable for instruments not in the table.
All keywords used for file selection are *required*.

========== ===========================================
Instrument  Keyword Attributes
========== ===========================================
WFI	    instrument, date, time
========== ===========================================
