"""Association Definitions: DMS-specific

Notes
-----
These associations are specifically defined for use in DMS.
"""

from romancal.associations.lib import rules_level2
from romancal.associations.registry import RegistryMarker

RegistryMarker.mark(rules_level2)
