.. _dark_selectors:

Reference Selection Keyword Attributes for DARK
+++++++++++++++++++++++++++++++++++++++++++++++
CRDS selects appropriate DARK references based on the following keyword attributes.
DARK is not applicable for instruments not in the table.

========== ===========================================
Instrument  Keyword Attributes
========== ===========================================
WFI	    instrument, detector, date, time
========== ===========================================
